# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

API_VERSION = 'api_version'
API_BASE_URL = 'api_base_url'
TABLEAU_BASE_URL = 'tableau_base_url'
SITE_NAME = 'site_name'
TABLEAU_ACCESS_TOKEN_NAME = 'tableau_personal_access_token_name'
TABLEAU_ACCESS_TOKEN_SECRET = 'tableau_personal_access_token_secret'
EXCLUDED_PROJECTS = 'excluded_projects'
EXTERNAL_CLUSTER_NAME = 'external_cluster_name'
EXTERNAL_SCHEMA_NAME = 'external_schema_name'
EXTERNAL_TABLE_TYPES = 'external_table_types'
CLUSTER = 'cluster'
DATABASE = 'database'
VERIFY_REQUEST = 'verify_request'
